import unittest
from unittest.mock import patch

from sohstationviewer.model.mseed_data.record_reader_helper import Unpacker
from sohstationviewer.model.reftek_data.reftek_reader.packet import \
    eh_et_payload_end_in_packet
from sohstationviewer.model.reftek_data.reftek_reader.packet_readers import (
    decode_uncompressed, decode_compressed, read_dt_packet, read_eh_et_packet,
    read_soh_packet,
)
from sohstationviewer.model.reftek_data.reftek_reader.packets import SOHExtendedHeader

unpacker = Unpacker('>')


class TestDecodeFunctions(unittest.TestCase):
    def setUp(self) -> None:
        self.header = b' ' * 24

    def test_decode_uncompressed_format_16(self):
        data_format = '16'
        with self.subTest('test_positive_number'):
            first_data_point_byte = b'\x06\x19'
            data_filler = b' ' * 998
            packet = self.header + first_data_point_byte + data_filler
            actual = decode_uncompressed(packet, data_format, unpacker)
            expected = 1561
            self.assertEqual(actual, expected)
        with self.subTest('test_negative_number'):
            first_data_point_byte = b'\xf9\xe4'
            data_filler = b' ' * 998
            packet = self.header + first_data_point_byte + data_filler
            actual = decode_uncompressed(packet, data_format, unpacker)
            expected = -1564
            self.assertEqual(actual, expected)

    def test_decode_uncompressed_format_32(self):
        data_format = '32'
        with self.subTest('test_positive_number'):
            first_data_point_byte = b'\x03\xc5\x4e\x9a'
            data_filler = b' ' * 996
            packet = self.header + first_data_point_byte + data_filler
            actual = decode_uncompressed(packet, data_format, unpacker)
            expected = 63262362
            self.assertEqual(actual, expected)
        with self.subTest('test_negative_number'):
            first_data_point_byte = b'\xf6\xac\xba\x00'
            data_filler = b' ' * 996
            packet = self.header + first_data_point_byte + data_filler
            actual = decode_uncompressed(packet, data_format, unpacker)
            expected = -156452352
            self.assertEqual(actual, expected)

    def test_decode_uncompressed_format_33(self):
        data_format = '33'
        with self.subTest('test_positive_number'):
            first_data_point_byte = b'\x03\xc5\x4e\x9a'
            data_filler = b' ' * 996
            packet = self.header + first_data_point_byte + data_filler
            actual = decode_uncompressed(packet, data_format, unpacker)
            expected = 63262362
            self.assertEqual(actual, expected)
        with self.subTest('test_negative_number'):
            first_data_point_byte = b'\xf6\xac\xba\x00'
            data_filler = b' ' * 996
            packet = self.header + first_data_point_byte + data_filler
            actual = decode_uncompressed(packet, data_format, unpacker)
            expected = -156452352
            self.assertEqual(actual, expected)

    def test_decode_compressed(self):
        data_format = 'C0'
        filler = b' ' * 40
        first_frame_code = b'\x00\x11\x11\x11'
        start_data_point_byte = b'0000'
        header = self.header + filler
        bytes_before_data = header + first_frame_code + start_data_point_byte
        with self.subTest('test_positive_number'):
            end_point_byte = b'\x03\xc5\x4e\x9a'
            data_filler = b' ' * 952
            packet = bytes_before_data + end_point_byte + data_filler
            actual = decode_compressed(packet, data_format, unpacker)
            expected = 63262362
            self.assertEqual(actual, expected)
        with self.subTest('test_negative_number'):
            end_point_byte = b'\xf6\xac\xba\x00'
            data_filler = b' ' * 952
            packet = bytes_before_data + end_point_byte + data_filler
            actual = decode_compressed(packet, data_format, unpacker)
            expected = -156452352
            self.assertEqual(actual, expected)


class TestReadDTPacket(unittest.TestCase):
    def setUp(self) -> None:
        self.header = b' ' * 16
        # We only test if the correct method is used to extract the data point,
        # so the data can be anything we want.
        self.data = b' ' * 1000

        uncompressed_patcher = patch(
            'sohstationviewer.model.reftek_data.reftek_reader.packet_readers.'
            'decode_uncompressed'
        )
        compressed_patcher = patch(
            'sohstationviewer.model.reftek_data.reftek_reader.packet_readers.'
            'decode_compressed'
        )
        self.mock_uncompressed = uncompressed_patcher.start()
        self.mock_compressed = compressed_patcher.start()
        self.addCleanup(uncompressed_patcher.stop)
        self.addCleanup(compressed_patcher.stop)

    def test_extended_header_is_extracted_correctly(self):
        extended_header_bytes = b'\x01\x11\x01\x02\x05\x00\x00\xc0'
        packet = self.header + extended_header_bytes + self.data
        extended_header, _ = read_dt_packet(packet, unpacker)
        self.assertEqual(extended_header.event_number, 111)
        self.assertEqual(extended_header.data_stream_number, 1)
        self.assertEqual(extended_header.channel_number, 2)
        self.assertEqual(extended_header.number_of_samples, 500)
        self.assertEqual(extended_header.flags, 0)
        self.assertEqual(extended_header.data_format, 'C0')

    def test_data_point_extracted_with_correct_method(self):
        with self.subTest('test_uncompressed_packet'):
            extended_header_bytes = b'\x01\x11\x01\x02\x05\x00\x00\x16'
            packet = self.header + extended_header_bytes + self.data
            read_dt_packet(packet, unpacker)
            self.assertTrue(self.mock_uncompressed.called)
            self.assertFalse(self.mock_compressed.called)

        self.mock_uncompressed.reset_mock()
        self.mock_compressed.reset_mock()

        with self.subTest('test_compressed_packet'):
            extended_header_bytes = b'\x01\x11\x01\x02\x05\x00\x00\xc0'
            packet = self.header + extended_header_bytes + self.data
            read_dt_packet(packet, unpacker)
            self.assertTrue(self.mock_compressed.called)
            self.assertFalse(self.mock_uncompressed.called)


class TestReadEHETPacket(unittest.TestCase):
    def setUp(self) -> None:
        header = b' ' * 16
        extended_header_bytes = b'\x01\x11\x01\x00\x00\x00\x00\xc0'
        # We only care about the length of the payload (the content is dealt
        # with somewhere else), and so it can contain dummy data.
        payload = b' ' * 1000
        self.packet = header + extended_header_bytes + payload

    def test_extended_header_is_extracted_correctly(self):
        extended_header, _ = read_eh_et_packet(self.packet, unpacker)
        self.assertEqual(extended_header.event_number, 111)
        self.assertEqual(extended_header.data_stream_number, 1)
        self.assertEqual(extended_header.channel_number, 0)
        self.assertEqual(extended_header.number_of_samples, 0)
        self.assertEqual(extended_header.flags, 0)
        self.assertEqual(extended_header.data_format, 'C0')

    def test_payload_extracted_correctly(self):
        _, payload = read_eh_et_packet(self.packet, unpacker)
        self.assertEqual(len(payload), eh_et_payload_end_in_packet - 24)


class TestReadSOHPacket(unittest.TestCase):
    """
    Test suite for packet_readers.read_soh_packet. We only test that the
    function has the correct interface, seeing as the intended purpose of this
    method is to be compatible with packet_readers.read_dt_packet and
    packet_readers.read_eh_et_packet interface-wise.
    """
    def test_correct_interface(self):
        packet = b' ' * 1024
        extended_header, payload = read_soh_packet(packet, unpacker)
        self.assertIsInstance(extended_header, SOHExtendedHeader)
        self.assertIsInstance(payload, bytes)

    def test_payload_has_correct_length(self):
        packet = b' ' * 1024
        extended_header, payload = read_soh_packet(packet, unpacker)
        self.assertEqual(len(payload), 1000)
